import base64
import hashlib
import hmac
import time
import uuid
from urllib.parse import quote, urlencode

from connector.generated import OAuth1Credential
from connector.oai.capability import Request, get_oauth_1


def create_oauth1_signature(
    method: str, url: str, params: dict[str, str], auth: OAuth1Credential
) -> str:
    """Generate HMAC-SHA256 OAuth 1.0 signature."""
    sorted_params = urlencode(sorted(params.items()), safe="~")
    base_string = "&".join([method.upper(), quote(url, safe=""), quote(sorted_params, safe="")])

    signing_key = f"{quote(auth.consumer_secret)}&{quote(auth.token_secret)}"
    signature = hmac.new(signing_key.encode(), base_string.encode(), hashlib.sha256).digest()
    return base64.b64encode(signature).decode()


def get_oauth1_headers(method: str, url: str, args: Request) -> dict[str, str]:
    """Generate OAuth 1.0 headers for HTTPX."""
    auth = get_oauth_1(args)

    nonce = str(uuid.uuid4())
    timestamp = str(int(time.time()))

    oauth_params = {
        "oauth_consumer_key": auth.consumer_key,
        "oauth_token": auth.token_id,
        "oauth_nonce": nonce,
        "oauth_timestamp": timestamp,
        "oauth_signature_method": "HMAC-SHA256",
        "oauth_version": "1.0",
    }

    oauth_signature = create_oauth1_signature(method, url, oauth_params, auth)
    oauth_params["oauth_signature"] = oauth_signature

    # Format the Authorization header
    auth_header = "OAuth " + ", ".join(f'{quote(k)}="{quote(v)}"' for k, v in oauth_params.items())
    return {"Authorization": auth_header}



import typing as t
from enum import Enum

import pydantic


class FieldType(str, Enum):
    SECRET = "SECRET"
    HIDDEN = "HIDDEN"
    MULTI_LINES = "MULTI_LINES"


def _extract_json_schema_extra(**kwargs) -> dict[str, t.Any]:
    json_schema_extra = (
        kwargs.pop("json_schema_extra") if "json_schema_extra" in kwargs else {}
    ) or {}
    return json_schema_extra


def SecretField(*args, **kwargs):
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_type"] = FieldType.SECRET
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)


def HiddenField(*args, **kwargs):
    """
    A field we don't want a user to see + fill out, but not a secret.
    """
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_type"] = FieldType.HIDDEN
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)


def MultiLinesField(*args, **kwargs):
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_type"] = FieldType.MULTI_LINES
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)


def GroupedField(group: str, *args, **kwargs):
    """
    A field that we want to group together in the UI

    :param group: The title of the group to group the field under
    """
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_group"] = group
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)


def IntegrationField(
    group: str | None = None,
    multiline: bool = False,
    secret: bool = False,
    *args,
    **kwargs,
):
    """
    A Field that can be used to combine multiple types of fields such as SecretField and GroupedField

    :param group: The title of the group to group the field under
    :param priority: The priority of the field, lower numbers are displayed first
    :param multiline: Whether the field should be a multi-line text field
    :param secret: Whether the field should be a secret field
    """
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    if group:
        json_schema_extra["x-field_group"] = group

    if all([multiline, secret]):
        raise ValueError("Cannot have both multiline and secret fields")
    if multiline:
        json_schema_extra["x-field_type"] = FieldType.MULTI_LINES
    if secret:
        json_schema_extra["x-field_type"] = FieldType.SECRET
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)



from pathlib import Path

import pytest
from connector.config import LOG_DIRECTORY_ENV_VAR, LOG_LEVEL_ENV_VAR, Config, LogLevel


@pytest.fixture
def clean_environment(monkeypatch):
    """Clear relevant environment variables before each test"""
    # monkeypatch.delenv removes env var if it exists, otherwise does nothing
    monkeypatch.delenv(LOG_LEVEL_ENV_VAR, raising=False)
    monkeypatch.delenv(LOG_DIRECTORY_ENV_VAR, raising=False)


def test_default_initialization(clean_environment):
    """Test that Config initializes with default values when no env vars are set"""
    config = Config()
    assert config.log_level == LogLevel.ERROR
    assert config.log_directory is None


def test_custom_log_level(monkeypatch):
    """Test that Config respects custom log level from environment"""
    monkeypatch.setenv(LOG_LEVEL_ENV_VAR, "DEBUG")
    config = Config()
    assert config.log_level == LogLevel.DEBUG


def test_custom_log_directory(monkeypatch):
    """Test that Config respects custom log directory from environment"""
    test_path = "/tmp/logs"
    monkeypatch.setenv(LOG_DIRECTORY_ENV_VAR, test_path)
    config = Config()
    assert config.log_directory == Path(test_path)


def test_invalid_log_level(monkeypatch):
    """Test that Config raises ValueError for invalid log level"""
    monkeypatch.setenv(LOG_LEVEL_ENV_VAR, "INVALID_LEVEL")
    with pytest.raises(ValueError) as exc_info:
        Config()
    assert "Invalid log level" in str(exc_info.value)



from unittest.mock import AsyncMock, patch
from urllib.parse import parse_qs, urlparse

import pytest
from connector.generated import (
    AuthModel,
    GetAuthorizationUrl,
    GetAuthorizationUrlRequest,
    HandleAuthorizationCallback,
    HandleAuthorizationCallbackRequest,
    HandleClientCredentials,
    HandleClientCredentialsRequest,
    OAuthClientCredential,
    OAuthCredential,
    RefreshAccessToken,
    RefreshAccessTokenRequest,
    StandardCapabilityName,
)
from connector.httpx_rewrite import AsyncClient
from connector.oai.capability import AuthRequest, get_settings
from connector.oai.errors import ConnectorError
from connector.oai.integration import DescriptionData, Integration
from connector.oai.modules.oauth_module import OAuthModule
from connector.oai.modules.oauth_module_types import (
    ClientAuthenticationMethod,
    OAuthConfig,
    OAuthFlowType,
    OAuthRequest,
    OAuthSettings,
    RequestDataType,
    RequestMethod,
)
from httpx import Request, Response
from pydantic import BaseModel, ValidationError


class Settings(BaseModel):
    subdomain: str


INTEGRATION_OAUTH_SETTINGS = OAuthSettings(
    authorization_url="https://example.com/auth",
    token_url="https://example.com/token",
    scopes={
        StandardCapabilityName.VALIDATE_CREDENTIALS: "test:scope another:scope",
        StandardCapabilityName.LIST_ACCOUNTS: "test:scope another:scope",
    },
)

INTEGRATION_OAUTH_SETTINGS_2 = OAuthSettings(
    authorization_url="https://example.com/auth2",
    token_url="https://example.com/token2",
    scopes={},
)

CLIENT_CREDENTIALS_INTEGRATION_OAUTH_SETTINGS = OAuthSettings(
    flow_type=OAuthFlowType.CLIENT_CREDENTIALS,
    token_url="https://example.com/token",
    scopes={
        StandardCapabilityName.VALIDATE_CREDENTIALS: "test:scope another:scope",
        StandardCapabilityName.LIST_ACCOUNTS: "test:scope another:scope",
    },
)


@pytest.fixture
def settings():
    return Settings(subdomain="test")


@pytest.fixture()
def integration():
    """
    FIXTURE: Create an integration with an OAuthModule
    """
    return Integration(
        app_id="test_app",
        auth=OAuthCredential,
        version="0.1.0",
        exception_handlers=[],
        settings_model=Settings,
        oauth_settings=INTEGRATION_OAUTH_SETTINGS,
        description_data=DescriptionData(user_friendly_name="test_oauth_module.py", categories=[]),
    )


@pytest.fixture()
def client_credentials_integration():
    """
    FIXTURE: Create an integration with an OAuthModule configured with a client credentials flow
    """
    return Integration(
        app_id="test_app",
        auth=OAuthClientCredential,
        version="0.1.0",
        exception_handlers=[],
        settings_model=Settings,
        oauth_settings=CLIENT_CREDENTIALS_INTEGRATION_OAUTH_SETTINGS,
        description_data=DescriptionData(user_friendly_name="test_oauth_module.py", categories=[]),
    )


@pytest.fixture()
def integration_with_credentials():
    """
    FIXTURE: Create an integration that uses OAuthModule through the credentials configs
    """
    return Integration(
        app_id="test_app",
        credentials=[
            OAuthConfig(
                id="test_credential",
                type=AuthModel.OAUTH,
                description="Test credential",
                oauth_settings=INTEGRATION_OAUTH_SETTINGS,
            ),
            OAuthConfig(
                id="test_credential_2",
                type=AuthModel.OAUTH,
                description="Test credential 2",
                oauth_settings=INTEGRATION_OAUTH_SETTINGS_2,
            ),
        ],
        version="0.1.0",
        exception_handlers=[],
        settings_model=Settings,
        description_data=DescriptionData(user_friendly_name="test_oauth_module.py", categories=[]),
    )


@pytest.fixture(autouse=True)
def cleanup_integration(integration, client_credentials_integration):
    """
    FIXTURE: Cleanup the integrations before each test is run
    """
    integration.oauth_settings = INTEGRATION_OAUTH_SETTINGS
    client_credentials_integration.oauth_settings = CLIENT_CREDENTIALS_INTEGRATION_OAUTH_SETTINGS

    integration.auth = OAuthCredential
    client_credentials_integration.auth = OAuthClientCredential
    integration_with_credentials.credentials = [
        OAuthConfig(
            id="test_credential",
            type=AuthModel.OAUTH,
            description="Test credential",
            oauth_settings=INTEGRATION_OAUTH_SETTINGS,
        ),
        OAuthConfig(
            id="test_credential_2",
            type=AuthModel.OAUTH,
            description="Test credential 2",
            oauth_settings=INTEGRATION_OAUTH_SETTINGS,
        ),
    ]

    yield


def get_oauth_module(integration: Integration) -> OAuthModule:
    """
    HELPER: Get the OAuthModule from the integration
    """
    for module in integration.modules:
        if isinstance(module, OAuthModule):
            return module
    raise ValueError("OAuthModule not found")


def test_get_scopes(integration):
    """
    Test if OAuthModule correctly returns the scopes defined in the integration settings
    """
    module = get_oauth_module(integration)

    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            client_id="test_client_id",
            redirect_uri="https://example.com/callback",
            state="test_state",
            scopes=[],
        ),
        settings={"subdomain": "test"},
    )

    if isinstance(module.settings[0], OAuthSettings):
        scopes = module._get_scopes(module.settings[0], request)
        assert set(scopes.split()) == {"test:scope", "another:scope"}
    else:
        with pytest.raises(ConnectorError):
            module._get_scopes(module.settings[0], request)  # type: ignore


def test_get_scopes_with_scopes_callable(integration):
    """
    Test if OAuthModule correctly returns the scopes defined in the integration settings if its a callable
    """

    def get_scopes(args: AuthRequest) -> dict[str, str]:
        return {
            StandardCapabilityName.VALIDATE_CREDENTIALS: "test:scope another:scope",
            StandardCapabilityName.LIST_ACCOUNTS: "test:scope another:scope",
        }

    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].scopes = get_scopes

    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            client_id="test_client_id",
            redirect_uri="https://example.com/callback",
            state="test_state",
            scopes=[],
        ),
        settings={"subdomain": "test"},
    )

    if isinstance(module.settings[0], OAuthSettings):
        scopes = module._get_scopes(module.settings[0], request)
        assert set(scopes.split()) == {"test:scope", "another:scope"}
    else:
        with pytest.raises(ConnectorError):
            module._get_scopes(module.settings[0], request)  # type: ignore


@pytest.mark.asyncio
async def test_dynamic_authorization_url(integration):
    """
    Test if OAuthModule correctly handles dynamic authorization URLs based on settings
    """

    def get_auth_url(args: AuthRequest) -> str:
        settings = get_settings(args, integration.settings_model)
        return f"https://{settings.subdomain}.example.com/auth"

    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].authorization_url = get_auth_url

    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            client_id="test_client_id",
            redirect_uri="https://test.example.com/callback",
            state="test_state",
            scopes=["scope1"],
        ),
        settings={"subdomain": "test"},
    )

    response = await integration.capabilities[StandardCapabilityName.GET_AUTHORIZATION_URL](request)

    parsed_url = urlparse(response.response.authorization_url)
    assert parsed_url.scheme == "https"
    assert parsed_url.netloc == "test.example.com"
    assert parsed_url.path == "/auth"


@pytest.mark.asyncio
async def test_dynamic_token_url(integration):
    """
    Test if OAuthModule correctly handles dynamic token URLs based on settings
    """

    def get_token_url(args: AuthRequest) -> str:
        settings = get_settings(args, integration.settings_model)
        return f"https://{settings.subdomain}.example.com/token"

    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].token_url = get_token_url

    request = HandleAuthorizationCallbackRequest(
        request=HandleAuthorizationCallback(
            client_id="test_client_id",
            client_secret="test_secret",
            redirect_uri_with_code="https://test.example.com/callback?code=test_code",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={"access_token": "test_token", "token_type": "bearer"},
        request=Request("POST", "https://test.example.com/token"),
    )

    with patch.object(AsyncClient, "request", return_value=mock_response) as mock_request:
        await integration.capabilities[StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK](
            request
        )

        mock_request.assert_called_once()
        assert mock_request.call_args.kwargs["url"] == "https://test.example.com/token"


@pytest.mark.asyncio
async def test_get_authorization_url(integration):
    """
    Test if OAuthModule correctly constructs the authorization URL
    """
    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            client_id="test_client_id",
            redirect_uri="https://test.example.com/callback",
            state="test_state",
            scopes=["scope1", "scope2", "scope3"],
        ),
        settings={"subdomain": "test"},
    )

    response = await integration.capabilities[StandardCapabilityName.GET_AUTHORIZATION_URL](request)

    parsed_url = urlparse(response.response.authorization_url)

    assert parsed_url.scheme == "https"
    assert parsed_url.netloc == "test.example.com"
    assert parsed_url.path == "/auth"

    assert "client_id=test_client_id" in parsed_url.query
    assert "redirect_uri=https://test.example.com/callback" in parsed_url.query
    assert "state=test_state" in parsed_url.query
    assert "scope=scope1 scope2 scope3" in parsed_url.query


@pytest.mark.asyncio
async def test_get_authorization_url_no_input_scopes(integration):
    """
    Test if OAuthModule falls back to the scopes defined in the integration settings
    if no scopes are provided in the request to the connector
    """
    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            client_id="test_client_id",
            redirect_uri="https://test.example.com/callback",
            state="test_state",
            scopes=[],
        ),
        settings={"subdomain": "test"},
    )

    response = await integration.capabilities[StandardCapabilityName.GET_AUTHORIZATION_URL](request)

    parsed_url = urlparse(response.response.authorization_url)
    scopes = parse_qs(parsed_url.query).get("scope", [""])[0].split()
    assert set(scopes) == {"another:scope", "test:scope"}


@pytest.mark.asyncio
async def test_handle_authorization_callback(integration):
    """
    Test if OAuthModule correctly handles the authorization callback
    """
    request = HandleAuthorizationCallbackRequest(
        request=HandleAuthorizationCallback(
            client_id="test_client_id",
            client_secret="test_client_secret",
            redirect_uri_with_code="https://test.example.com/callback?code=test_code",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
            "refresh_token": "test_refresh_token",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    with patch.object(AsyncClient, "request", return_value=mock_response):
        response = await integration.capabilities[
            StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK
        ](request)

    assert response.response.access_token == "test_access_token"
    assert response.response.token_type == "bearer"
    assert response.response.refresh_token == "test_refresh_token"


@pytest.mark.asyncio
async def test_handle_client_credentials_request(client_credentials_integration):
    """
    Test if OAuthModule correctly handles the client credentials request
    """
    request = HandleClientCredentialsRequest(
        request=HandleClientCredentials(
            client_id="test_client_id",
            client_secret="test_client_secret",
            scopes=["scope1", "scope2", "scope3"],
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    with patch.object(AsyncClient, "request", return_value=mock_response):
        response = await client_credentials_integration.capabilities[
            StandardCapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST
        ](request)

    assert response.response.access_token == "test_access_token"
    assert response.response.token_type == "bearer"


@pytest.mark.asyncio
async def test_refresh_access_token(integration):
    """
    Test if OAuthModule correctly refreshes the access token
    """
    request = RefreshAccessTokenRequest(
        request=RefreshAccessToken(
            client_id="test_client_id",
            client_secret="test_client_secret",
            refresh_token="test_refresh_token",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "new_access_token",
            "token_type": "Bearer",
            "expires_in": 3600,
            "refresh_token": "new_refresh_token",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    with patch.object(AsyncClient, "request", return_value=mock_response):
        response = await integration.capabilities[StandardCapabilityName.REFRESH_ACCESS_TOKEN](
            request
        )

    assert response.response.access_token == "new_access_token"
    assert response.response.token_type == "bearer"
    assert response.response.refresh_token == "new_refresh_token"


@pytest.mark.asyncio
async def test_send_authorized_request_client_secret_basic(integration):
    """
    Test if OAuthModule correctly constructs the request for the token endpoint
    when using client secret basic authentication (header auth)
    """
    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].client_auth = ClientAuthenticationMethod.CLIENT_SECRET_BASIC

    request = HandleAuthorizationCallbackRequest(
        request=HandleAuthorizationCallback(
            client_id="test_client_id",
            client_secret="test_client_secret",
            redirect_uri_with_code="https://test.example.com/callback?code=test_code",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
            "refresh_token": "test_refresh_token",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    mock_client = AsyncMock(spec=AsyncClient)
    mock_client.request.return_value = mock_response

    await module._send_authorized_request(
        "https://test.example.com/token", "authorization_code", mock_client, request
    )

    mock_client.request.assert_called_once()
    call_args = mock_client.request.call_args
    assert call_args.kwargs["auth"] is not None
    assert "client_id" not in call_args.kwargs["data"]
    assert "client_secret" not in call_args.kwargs["data"]


@pytest.mark.asyncio
async def test_send_authorized_request_query_request_type(integration):
    """
    Test if OAuthModule correctly constructs the request for the token endpoint
    when using a request type with QUERY as the data parameter
    """
    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].request_type = OAuthRequest(
            method=RequestMethod.GET, data=RequestDataType.QUERY
        )

    request = RefreshAccessTokenRequest(
        request=RefreshAccessToken(
            client_id="test_client_id",
            client_secret="test_client_secret",
            refresh_token="test_refresh_token",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
            "refresh_token": "test_refresh_token",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    mock_client = AsyncMock(spec=AsyncClient)
    mock_client.request.return_value = mock_response

    await module._send_authorized_request(
        "https://test.example.com/token", "refresh_token", mock_client, request
    )

    mock_client.request.assert_called_once()
    call_args = mock_client.request.call_args
    assert call_args.kwargs["method"] == "GET"
    assert call_args.kwargs["params"] is not None
    assert call_args.kwargs["data"] is None
    assert call_args.kwargs["json"] is None


@pytest.mark.asyncio
async def test_send_authorized_request_json_request_type(integration):
    """
    Test if OAuthModule correctly constructs the request for the token endpoint
    when using a request type with JSON as the data parameter
    """
    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].request_type = OAuthRequest(
            method=RequestMethod.GET, data=RequestDataType.JSON
        )

    request = RefreshAccessTokenRequest(
        request=RefreshAccessToken(
            client_id="test_client_id",
            client_secret="test_client_secret",
            refresh_token="test_refresh_token",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
            "refresh_token": "test_refresh_token",
        },
        request=Request("GET", "https://test.example.com/token"),
    )

    mock_client = AsyncMock(spec=AsyncClient)
    mock_client.request.return_value = mock_response

    await module._send_authorized_request(
        "https://test.example.com/token", "refresh_token", mock_client, request
    )

    mock_client.request.assert_called_once()
    call_args = mock_client.request.call_args
    assert call_args.kwargs["method"] == "GET"
    assert call_args.kwargs["params"] is None
    assert call_args.kwargs["data"] is None
    assert call_args.kwargs["json"] is not None


@pytest.mark.asyncio
async def test_send_authorized_request_formdata_request_type(integration):
    """
    Test if OAuthModule correctly constructs the request for the token endpoint
    when using a request type with FORMDATA as the data parameter
    """
    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].request_type = OAuthRequest(
            method=RequestMethod.GET, data=RequestDataType.FORMDATA
        )

    request = RefreshAccessTokenRequest(
        request=RefreshAccessToken(
            client_id="test_client_id",
            client_secret="test_client_secret",
            refresh_token="test_refresh_token",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
            "refresh_token": "test_refresh_token",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    mock_client = AsyncMock(spec=AsyncClient)
    mock_client.request.return_value = mock_response

    await module._send_authorized_request(
        "https://test.example.com/token", "refresh_token", mock_client, request
    )

    mock_client.request.assert_called_once()
    call_args = mock_client.request.call_args
    assert call_args.kwargs["method"] == "GET"
    assert call_args.kwargs["params"] is None
    assert call_args.kwargs["data"] is not None
    assert call_args.kwargs["json"] is None


@pytest.mark.asyncio
async def test_handle_authorization_callback_error(integration):
    """
    Test if OAuthModule correctly handles errors during authorization callback
    """
    request = HandleAuthorizationCallbackRequest(
        request=HandleAuthorizationCallback(
            client_id="test_client_id",
            client_secret="test_client_secret",
            redirect_uri_with_code="https://test.example.com/callback?code=test_code",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        400,
        json={"error": "invalid_grant", "error_description": "Invalid authorization code"},
        request=Request("POST", "https://test.example.com/token"),
    )

    with patch.object(AsyncClient, "request", return_value=mock_response):
        with pytest.raises(ConnectorError):
            await integration.capabilities[StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK](
                request
            )


@pytest.mark.asyncio
async def test_invalid_grant_type(integration):
    """
    Test if OAuthModule correctly handles an invalid grant type
    """
    module = get_oauth_module(integration)
    request = RefreshAccessTokenRequest(
        request=RefreshAccessToken(
            client_id="test_client_id",
            client_secret="test_client_secret",
            refresh_token="test_refresh_token",
        ),
        settings={"subdomain": "test"},
    )

    mock_client = AsyncMock(spec=AsyncClient)
    with pytest.raises(ValueError, match="Unsupported grant_type"):
        await module._send_authorized_request(
            "https://test.example.com/token", "invalid_grant_type", mock_client, request
        )


@pytest.mark.asyncio
async def test_oauth_credentials_conversion_error(integration):
    """
    Test if OAuthModule correctly handles errors when converting raw JSON to OAuthCredentials
    """
    request = HandleAuthorizationCallbackRequest(
        request=HandleAuthorizationCallback(
            client_id="test_client_id",
            client_secret="test_client_secret",
            redirect_uri_with_code="https://test.example.com/callback?code=test_code",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            # Missing required fields
            "some_other_field": "value",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    with patch.object(AsyncClient, "request", return_value=mock_response):
        with pytest.raises(ValidationError, match="1 validation error for OauthCredentials"):
            await integration.capabilities[StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK](
                request
            )


@pytest.mark.asyncio
async def test_handle_authorization_callback_pkce(integration):
    """
    Test if OAuthModule correctly handles the authorization callback with PKCE
    """
    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].client_auth = ClientAuthenticationMethod.CLIENT_SECRET_POST
        module.settings[0].pkce = True

    request = HandleAuthorizationCallbackRequest(
        request=HandleAuthorizationCallback(
            client_id="test_client_id",
            client_secret="test_client_secret",
            redirect_uri_with_code="https://test.example.com/callback?code=test_code",
            code_verifier="test_code_verifier",
        ),
        settings={"subdomain": "test"},
    )

    mock_response = Response(
        200,
        json={
            "access_token": "test_access_token",
            "token_type": "Bearer",
            "refresh_token": "test_refresh_token",
        },
        request=Request("POST", "https://test.example.com/token"),
    )

    mock_client = AsyncMock(spec=AsyncClient)
    mock_client.request.return_value = mock_response

    await module._send_authorized_request(
        "https://test.example.com/token", "authorization_code", mock_client, request
    )

    mock_client.request.assert_called_once()
    call_args = mock_client.request.call_args
    assert call_args.kwargs["data"] == {
        "grant_type": "authorization_code",
        "code": "test_code",
        "redirect_uri": "https://test.example.com/callback",
        "code_verifier": "test_code_verifier",  # must be included
        "client_id": "test_client_id",
        "client_secret": "test_client_secret",
    }


@pytest.mark.asyncio
async def test_get_authorization_url_pkce(integration):
    """
    Test if OAuthModule correctly handles the authorization URL with PKCE
    """
    module = get_oauth_module(integration)
    if isinstance(module.settings[0], OAuthSettings):
        module.settings[0].pkce = True

    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            client_id="test_client_id",
            redirect_uri="https://test.example.com/callback",
            state="test_state",
            scopes=["scope1"],
        ),
        settings={"subdomain": "test"},
    )

    response = await integration.capabilities[StandardCapabilityName.GET_AUTHORIZATION_URL](request)

    parsed_url = urlparse(response.response.authorization_url)
    query_params = parse_qs(parsed_url.query)
    assert "code_challenge" in query_params
    assert "code_challenge_method" in query_params


@pytest.mark.asyncio
async def test_check_oauth_module_settings_when_using_credentials(integration_with_credentials):
    """
    Test if OAuthModule correctly collects the settings from the credentials configs
    """
    module = get_oauth_module(integration_with_credentials)

    assert isinstance(module.settings[0], OAuthConfig)
    assert isinstance(module.settings[1], OAuthConfig)

    assert module.settings[0].id == "test_credential"
    assert module.settings[1].id == "test_credential_2"


@pytest.mark.asyncio
async def test_get_authorization_url_with_credentials(integration_with_credentials):
    """
    Test if OAuthModule correctly handles the authorization URL when configured with credentials
    """
    module = get_oauth_module(integration_with_credentials)
    assert not isinstance(module.settings[0], OAuthSettings)
    assert len(module.settings) == 2

    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            credential_id="test_credential",  # match
            client_id="test_client_id",
            redirect_uri="https://test.example.com/callback",
            state="test_state",
            scopes=["scope1", "scope2", "scope3"],
        ),
        settings={"subdomain": "test"},
    )

    response = await integration_with_credentials.capabilities[
        StandardCapabilityName.GET_AUTHORIZATION_URL
    ](request)

    parsed_url = urlparse(response.response.authorization_url)
    assert parsed_url.path == "/auth"

    assert "client_id=test_client_id" in parsed_url.query
    assert "redirect_uri=https://test.example.com/callback" in parsed_url.query
    assert "state=test_state" in parsed_url.query
    assert "scope=scope1 scope2 scope3" in parsed_url.query


@pytest.mark.asyncio
async def test_get_authorization_url_with_credentials_2(integration_with_credentials):
    """
    Test if OAuthModule correctly handles the authorization URL when configured with credentials
    """
    request = GetAuthorizationUrlRequest(
        request=GetAuthorizationUrl(
            credential_id="test_credential_2",  # match
            client_id="test_client_id",
            redirect_uri="https://test.example.com/callback",
            state="test_state",
            scopes=["scope1", "scope2", "scope3"],
        ),
        settings={"subdomain": "test"},
    )

    response = await integration_with_credentials.capabilities[
        StandardCapabilityName.GET_AUTHORIZATION_URL
    ](request)

    parsed_url = urlparse(response.response.authorization_url)
    assert parsed_url.path == "/auth2"  # parent credential settings are changed



"""Cases for testing ``unassign_entitlement`` capability."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    UnassignedEntitlement,
    UnassignEntitlement,
    UnassignEntitlementRequest,
    UnassignEntitlementResponse,
)
from connector.utils.test import http_error_message

from tests.common_mock_data import INVALID_AUTH, SETTINGS, VALID_AUTH
from connector.tests.type_definitions import MockedResponse, ResponseBodyMap

from connector.generated.models.standard_capability_name import StandardCapabilityName

Case: t.TypeAlias = tuple[
    StandardCapabilityName,
    UnassignEntitlementRequest,
    ResponseBodyMap,
    UnassignEntitlementResponse | ErrorResponse,
]

# repeat following casess for all entitlements

def case_assign_entitlement_1_404() -> Case:
    """Authorized request for non-existing entitlement should fail."""
    args = UnassignEntitlementRequest(
        request=UnassignEntitlement(
            account_integration_specific_id="",
            resource_integration_specific_id="",
            resource_type="",
            entitlement_integration_specific_id="",
            entitlement_type="",
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                "",
                404,
            ),
            status_code=httpx.codes.NOT_FOUND,
            error_code=ErrorCode.NOT_FOUND,
            app_id="{name}",
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:unassign_entitlement",
        ),
    )
    return StandardCapabilityName.UNASSIGN_ENTITLEMENT, args, response_body_map, expected_response


def case_unassign_entitlement_1_200() -> Case:
    """Successfully unassign entitlement."""
    args = UnassignEntitlementRequest(
        request=UnassignEntitlement(
            account_integration_specific_id="",
            resource_integration_specific_id="",
            resource_type="",
            entitlement_integration_specific_id="",
            entitlement_type="",
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = UnassignEntitlementResponse(
        response=UnassignedEntitlement(unassigned=True),
    )
    return StandardCapabilityName.UNASSIGN_ENTITLEMENT, args, response_body_map, expected_response



import json
import typing as t

import httpx
from connector.generated import (
    BasicCredential,
    Error,
    ErrorCode,
    ErrorResponse,
    ListAccountsRequest,
    ListAccountsResponse,
    StandardCapabilityName,
)
from connector.oai.errors import HTTPHandler
from connector.oai.integration import DescriptionData, Integration

Case = tuple[
    Integration,
    StandardCapabilityName,
    str,
    dict[str, t.Any],
]


def case_http_status_error() -> Case:
    """Test if HTTPStatusError can be handled with HTTPHandler.

    We register capability that is mocked to raise ``HTTPStatusError``.
    Since the integration has ``HTTPHandler`` registered for handling
    such error, we should end up with ``ErrorResponse`` that contains
    the details about HTTP error.
    """
    app_id = "test"
    integration = Integration(
        app_id=app_id,
        version="0.1.0",
        auth=BasicCredential,
        exception_handlers=[
            (httpx.HTTPStatusError, HTTPHandler, None),
        ],
        description_data=DescriptionData(user_friendly_name="hi, testing", categories=[]),
    )
    # will be mocked with actual response just to avoid making request
    requested_url = "https://httpstat.us/401"
    response_status_code = httpx.codes.UNAUTHORIZED

    capability_name = StandardCapabilityName.LIST_ACCOUNTS

    @integration.register_capability(capability_name)
    async def list_accounts(args: ListAccountsRequest) -> ListAccountsResponse:
        def request_handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(
                text="401 Unauthorized",
                status_code=response_status_code,
            )

        with httpx.Client(transport=httpx.MockTransport(request_handler)) as client:
            _response_text = client.get(requested_url).raise_for_status().text

        # this should never happen
        return ListAccountsResponse(
            response=[],
            raw_data=None,
        )

    request_data = json.dumps(
        {
            "auth": {"basic": {"username": "user", "password": "pass"}},
            "request": {},
        }
    )
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message="[401][https://httpstat.us/401] 401 Unauthorized",
            error_code=ErrorCode.UNAUTHORIZED,
            app_id=app_id,
            status_code=response_status_code,
            raised_by="HTTPStatusError",
            raised_in=f"{__name__}:{capability_name.value}",
        ),
    )
    return (
        integration,
        StandardCapabilityName.LIST_ACCOUNTS,
        request_data,
        expected_response.model_dump(),
    )



from urllib.parse import parse_qs, urljoin, urlparse

from connector.generated.models.handle_authorization_callback_request import (
    HandleAuthorizationCallbackRequest,
)


def parse_auth_code_and_redirect_uri(args: HandleAuthorizationCallbackRequest):
    redirect_uri_with_code = args.request.redirect_uri_with_code
    parsed_uri = urlparse(redirect_uri_with_code)
    base_url = f"{parsed_uri.scheme}://{parsed_uri.netloc}"
    path = parsed_uri.path
    original_redirect_uri = urljoin(base_url, path)
    query_params = parse_qs(parsed_uri.query)
    authorization_code = query_params.get("code", [None])[0]

    return authorization_code, original_redirect_uri



import os
import re
from argparse import Namespace, ArgumentParser, _SubParsersAction
from pathlib import Path

TEMPLATE_DIR = Path(__file__).parent / "templates"


def snakecase(name: str) -> str:
    return name.lower().replace(" ", "_").replace("-", "_")


def hyphenate(name: str) -> str:
    return name.lower().replace(" ", "-").replace("_", "-")


def namecase(name: str) -> str:
    return name.replace("_", " ").replace("-", " ").title()


def pascalcase(name: str) -> str:
    parts = [part.title() for part in name.replace("_", "-").split("-")]
    return "".join(parts)


class ScaffoldError(Exception):
    pass


IGNORE_FILE_REGEX = re.compile("\\.pytest_cache|\\.pyc$")

def setup_args(scaffold_parser: ArgumentParser) -> None:
    scaffold_parser.add_argument("name", help="Name of the new connector")
    scaffold_parser.add_argument(
        "directory", type=Path, help="Directory to create the connector in"
    )
    scaffold_parser.add_argument("--force-overwrite", "-f", action="store_true")
    scaffold_parser.add_argument("--tests-only", "-t", action="store_true")
    scaffold_parser.add_argument('--author-name', help="Author name for the generated package", type=str)
    scaffold_parser.add_argument('--author-email', help="Author email for the generated package", type=str)


def scaffold(args: Namespace):
    directory = args.directory
    name = args.name
    author = args.author_name or input("Author: ")
    author_email = args.author_email or input("Author email: ")

    connector_dir = Path(directory)
    try:
        next(connector_dir.iterdir())
    except (StopIteration, FileNotFoundError):
        pass
    else:
        if not args.force_overwrite:
            raise FileExistsError(f"Directory {connector_dir} is not empty, use --force-overwrite if you wish to overwrite it!")
    connector_dir.mkdir(parents=True, exist_ok=True)

    # Copy template files to the new connector directory
    output_paths = []
    try:
        for template_file in TEMPLATE_DIR.rglob("*"):
            if template_file.is_file() and not IGNORE_FILE_REGEX.search(str(template_file)):
                # Optionally only overwrite tests
                is_test_file = "test" in str(template_file)
                if args.tests_only and not is_test_file:
                    continue

                with template_file.open() as f:
                    try:
                        content = f.read()
                    except UnicodeDecodeError:
                        raise TypeError(f"Non-unicode file at {str(template_file)}") from None
                print(f"scaffolding:{template_file}")
                content = content.format(
                    author=author,
                    author_email=author_email,
                    name=snakecase(name),
                    title=namecase(name),
                    pascal=pascalcase(name),
                    hyphenated_name=hyphenate(name),
                )
                # Preserve the relative path of the template file
                relative_path = template_file.relative_to(TEMPLATE_DIR)
                output_path = connector_dir / relative_path
                output_path = Path(str(output_path).replace("connector_name_here", snakecase(name)))
                output_path.parent.mkdir(parents=True, exist_ok=True)
                if os.path.isfile(output_path) and not args.force_overwrite:
                    raise ScaffoldError(
                        f"Refusing to overwrite without --force-overwrite: {output_path}"
                    )
                output_paths.append(output_path)
                with output_path.open("w") as f:
                    f.write(content)
    except Exception as e:
        for output_path in output_paths:
            output_path.unlink()
        raise ScaffoldError(f"Failed to scaffold connector: {e}") from e



from pytest_httpx import HTTPXMock

from connector.tests.type_definitions import ResponseBodyMap


def mock_requests(
    response_body_map: ResponseBodyMap, httpx_mock: HTTPXMock, *, host: str | None = None
):
    if not response_body_map:
        # Don't mock any requests, and use the default behavior of
        # httpx_mock which is HTTP 200, empty body
        return

    for method, requests in response_body_map.items():
        for request_line, response in requests.items():
            if isinstance(response.response_body, str):
                httpx_mock.add_response(
                    method=method,
                    url=f"{host or ''}{request_line}",
                    content=response.response_body.encode(),
                    status_code=response.status_code,
                )
            else:
                httpx_mock.add_response(
                    method=method,
                    url=f"{host or ''}{request_line}",
                    json=response.response_body,
                    status_code=response.status_code,
                    headers=response.headers if response.headers else None,
                    match_json=response.request_json_body if response.request_json_body else None,
                )



import asyncio
import gzip
import json
import logging
import os
import subprocess
import sys
from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter
from datetime import datetime, timezone

from connector.httpx_rewrite import proxy_settings
from connector.logging import set_logger_config
from connector.oai.integration import Integration
from connector.pydantic import get_pydantic_model
from connector.utils import proxy_utils

logger = logging.getLogger("integration-connectors.sdk")

# Hacking commands
# ----------------


def _prep_hacking_command(args: Namespace):
    data = vars(args)
    data.pop("command")
    data.pop("func")
    return data


def http_integration_server(
    integration: Integration, port: int = 8000, reload: bool = False, use_proxy: bool = False
):
    from connector.http_server import runserver

    try:
        runserver(port=port, reload=reload, integration=integration, use_proxy=use_proxy)
    except KeyboardInterrupt:
        pass


def build_executable(path: str) -> None:
    try:
        subprocess.run(["pyinstaller", "--version"], check=True)
    except FileNotFoundError:
        print("PyInstaller not found in PATH. Please pip install pyinstaller")
        return

    command = [
        "pyinstaller",
        path,
        "--noconsole",
        "--onefile",
        "--clean",
        "--paths=projects/libs/python/connector-sdk",
    ]
    if __file__ not in "site-packages":
        command.append("--paths=projects/libs/python/connector-sdk")
    subprocess.run(command)


def create_integration_hacking_parser(integration: Integration, parser: ArgumentParser) -> None:
    subparsers = parser.add_subparsers(dest="command")

    http_server_parser = subparsers.add_parser(
        "http-server",
        help="Run this connector as an HTTP server.",
        description="\n".join(
            [
                "Run this connector as an HTTP server.",
                " - You can call capabilities via POST /<capability name> with the input JSON as the request body.",
                " - API docs are at /docs",
                " - The OpenAPI spec is at /openapi.json",
            ]
        ),
        formatter_class=RawDescriptionHelpFormatter,
    )
    http_server_parser.add_argument(
        "--port", "-p", type=int, default=8000, help="The port to run the server on."
    )
    http_server_parser.add_argument(
        "--reload", action="store_true", help="Enable hot reload for the server."
    )
    http_server_parser.add_argument(
        "--use-proxy", "--sip", action="store_true", help="Use the Lumos SIP to proxy requests."
    )
    http_server_parser.set_defaults(
        func=lambda args: http_integration_server(integration, **_prep_hacking_command(args))
    )

    build_executable_parser = subparsers.add_parser(
        "build-executable",
        help=(
            "Create a single file executable with PyInstaller. Provide the path to your library's"
            " main.py file."
        ),
    )
    build_executable_parser.add_argument("path", type=str, help="The path to the main.py file.")
    build_executable_parser.set_defaults(
        func=lambda args: build_executable(**_prep_hacking_command(args))
    )

    return None


# Actual Commands
# ---------------


def get_result_file_path(args: Namespace) -> str | None:
    if not args.result_file_path:
        return None
    result_file_path = args.result_file_path.strip('"').strip("'")
    if not result_file_path.endswith(".gz"):
        raise ValueError("The result file name must end with the .gz extension.")
    return result_file_path


def capability_executor(integration: Integration, args: Namespace):
    """Executes a command from the CLI."""
    # validate that a valid gzip file name was provided
    result_file_path = get_result_file_path(args)

    if args.command == "info":
        output = json.dumps(integration.info().model_dump(), sort_keys=True)
    else:
        if args.use_proxy:
            with proxy_settings(
                proxy_url=proxy_utils.get_proxy_url(),
                proxy_headers={
                    "X-Lumos-Proxy-Auth": proxy_utils.get_proxy_token_sync().token,
                },
            ):
                output = asyncio.run(integration.dispatch(args.command, args.json))
        else:
            output = asyncio.run(integration.dispatch(args.command, args.json))

    if result_file_path:
        logger.info(f"Attempting to open file name: {result_file_path}")
        with open(result_file_path, "w") as result_file:
            logger.info(f"File opened to write: {result_file_path}")
            logger.info("compressing result")
            output_bytes = output.encode("utf-8")
            gzipped_result = gzip.compress(output_bytes)
            logger.info("writing compressed result to file")
            result_file.buffer.write(gzipped_result)
            logger.info("compressed result written to file")
        logger.info(f"Result saved to {args.result_file_path}")
    else:
        logger.info("Result printing to console")
        print(output)

    logger.info("Command completed")


CAPABILITY_PREFIX = "\n      "


def collect_capabilities(integration: Integration, no_print: bool = False) -> ArgumentParser:
    """
    Collect all methods from an Integration class and create a CLI
    command for each.
    """
    executed = os.path.basename(sys.argv[0])
    capability_helps: list[str] = []
    capability_helps = sorted(integration.capabilities.keys())
    parser = ArgumentParser(
        description=f"Lumos integration CLI for {integration.description_data.user_friendly_name}",
        usage=f"""{executed} CAPABILITY [--json JSON STRING] [--result-file-path FILEPATH]

    Examples:

    {executed} info
        Print the Info schema for how to call this connector, and what it supports. This is the only
        capability that takes no arguments.

    {executed} validate_credentials --json '{'{}'}'
        Check if you're passing enough auth credentials and settings to connect to the underlying
        app tenant.

    All capabilities except 'info' require a JSON argument.

    A typical JSON argument looks like

    {'{'}
        "auth": {'{...}'},
        "settings": {'{...}'},
        "request": {'{...}'}
    {'}'}

    All capabilities:

{CAPABILITY_PREFIX}{CAPABILITY_PREFIX.join([c for c in capability_helps])}

""",
    )

    subparsers = parser.add_subparsers(dest="command")

    subparser = subparsers.add_parser("info", description=integration.info.__doc__)
    _add_arguments_to_subparser(subparser)
    subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    for capability_name, capability in integration.capabilities.items():
        subparser = subparsers.add_parser(capability_name, description=capability.__doc__)

        try:
            get_pydantic_model(capability.__annotations__)
        except ValueError:
            pass
        else:
            subparser.add_argument("--json", type=str, help="JSON input", required=True)
            _add_arguments_to_subparser(subparser)

        subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    hacking_subparser = subparsers.add_parser("hacking")
    create_integration_hacking_parser(integration, hacking_subparser)

    return parser


def _add_arguments_to_subparser(subparser: ArgumentParser):
    subparser.add_argument(
        "--result-file-path",
        type=str,
        help=(
            "The path to the file to save the result to. If this file already exists, "
            "its content will be overwritten. If not provided, the result will be "
            "printed to the console. The result file will be gzip compressed so the "
            "input file name should end with the .gz extension."
        ),
        default=None,
    )
    subparser.add_argument(
        "--use-proxy",
        "--sip",
        action="store_true",
        help="Use the Lumos SIP to proxy requests. This should not be used by non-Lumos developers.",
    )


def run_integration(
    integration: Integration,
    no_print: bool = False,
) -> None:
    set_logger_config(integration.app_id)
    logger.info("Running command started at %s", datetime.now(timezone.utc))
    try:
        """Run a command from the CLI, integration version."""
        parser = collect_capabilities(integration, no_print)
        args = parser.parse_args()
        logger.info("Command arguments: %s", args)
        if not args.command:
            print("No command passed in", file=sys.stderr)
            parser.print_help(file=sys.stderr)
            sys.exit(1)
        args.func(args)
    except Exception as e:
        logger.error(
            f"Error running command exception class: {e.__class__.__name__} exception: {e}"
        )
        logger.error("Stack trace:", exc_info=True)
        raise e
    logger.info("Command completed at %s", datetime.now(timezone.utc))



import typing as t

import httpx
import pydantic

DTO = t.TypeVar("DTO", bound=pydantic.BaseModel)


def create_client_response(
    response: httpx.Response,
    dto: type[DTO],
) -> DTO:
    response.raise_for_status()
    response_body = response.json()
    if isinstance(response_body, list):
        return dto(root=response_body)
    else:
        return dto(**response_body)



from abc import abstractmethod
from typing import TYPE_CHECKING

from connector.generated import StandardCapabilityName

if TYPE_CHECKING:
    from connector.oai.integration import Integration


class BaseIntegrationModule:
    """
    Base class for all integration modules.
    Integration modules allow you to register "global" capabilities that are not specific to a particular integration.
    """

    capabilities: list[StandardCapabilityName | str]

    def __init__(self):
        self.capabilities = []

    def add_capability(self, capability: str):
        """Add a capability to the module."""
        self.capabilities.append(capability)

    def get_capability(self, capability: str) -> StandardCapabilityName | str | None:
        """Get a capability from the module."""
        for cap in self.capabilities:
            if cap == capability:
                return cap
        return None

    @abstractmethod
    def register(self, integration: "Integration"):
        """Register all capabilities of the module / the module."""
        pass



from enum import Enum

from connector.generated import EntitlementType, ResourceType


class {pascal}ResourceTypes(str, Enum):
    GLOBAL_RESOURCE = "GLOBAL_RESOURCE"

class {pascal}EntitlementTypes(str, Enum):
    GLOBAL_ROLE = "ROLE"

resource_types: list[ResourceType] = [
    ResourceType(
        type_id={pascal}ResourceTypes.GLOBAL_RESOURCE,
        type_label="Global Resource",
    )
]

entitlement_types: list[EntitlementType] = [
    EntitlementType(
        type_id={pascal}EntitlementTypes.GLOBAL_ROLE,
        type_label="Role",
        resource_type_id="",
        min=0,
        # You can also set a max, if users can't have infinite of these entitlements
        # max=1,
    )
]



import typing as t

from connector.generated import (
    BasicCredential,
    StandardCapabilityName,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from connector.oai.integration import DescriptionData, Integration
from connector.oai.modules.base_module import BaseIntegrationModule

Case: t.TypeAlias = tuple[
    Integration,
    BaseIntegrationModule,
]


def case_add_module() -> Case:
    integration = Integration(
        app_id="test",
        version="0.1.0",
        auth=BasicCredential,
        exception_handlers=[],
        description_data=DescriptionData(
            user_friendly_name="test_modules_cases.py",
            categories=[],
        ),
    )

    class TestModule(BaseIntegrationModule):
        def register(self, integration: Integration):
            print("registered module")

    return integration, TestModule()


def case_register_module_capabilities() -> Case:
    integration = Integration(
        app_id="test",
        version="0.1.0",
        auth=BasicCredential,
        exception_handlers=[],
        description_data=DescriptionData(
            user_friendly_name="test_modules_cases.py",
            categories=[],
        ),
    )

    class TestModuleWithCapabilities(BaseIntegrationModule):
        def register(self, integration: Integration):
            self.integration = integration
            self.register_capability()
            self.add_capability(StandardCapabilityName.VALIDATE_CREDENTIALS)

        def register_capability(self):
            @self.integration.register_capability(StandardCapabilityName.VALIDATE_CREDENTIALS)
            async def validate_credentials(
                args: ValidateCredentialsRequest,
            ) -> ValidateCredentialsResponse:
                return ValidateCredentialsResponse(
                    response=ValidatedCredentials(
                        valid=True,
                        unique_tenant_id="test",
                    )
                )

    return integration, TestModuleWithCapabilities()



"""Test cases for test_pagination_utils.

Those cases cover scenarios for ``encode_pagination`` function.
"""

import typing as t

from tests.utils.test_pagination import Pagination

Case: t.TypeAlias = list[Pagination]


def case_single_url_page() -> Case:
    """Pagination for one call with pagination."""
    paginations = [
        Pagination(endpoint="/users", page=1),
    ]
    return paginations


def case_two_url_page() -> Case:
    """Pagination for one call with pagination."""
    paginations = [
        Pagination(endpoint="/users", page=1),
        Pagination(endpoint="/roles", page=20),
    ]
    return paginations



import argparse
import os.path
from pathlib import Path

from connector.__about__ import __version__
from connector.compile import DEFAULT_EXCLUDE_MODULES, compile_executable_for_onprem
from connector.scaffold.create import scaffold, setup_args as setup_scaffold_args


def main():
    parser = argparse.ArgumentParser(description="Lumos Connectors CLI")
    parser.add_argument(
        "--version", "-v", help="Print the version of this library and exit", action="store_true"
    )

    command_subparsers = parser.add_subparsers(dest="command")

    scaffold_parser = command_subparsers.add_parser("scaffold", help="Create a new connector")
    setup_scaffold_args(scaffold_parser)

    command_subparsers.add_parser("spec", help="Print the OpenAPI spec")

    compile_parser = command_subparsers.add_parser(
        "compile-on-prem",
        description="Compile a Python connector for on-prem use",
        help="Compile a Python connector for on-prem use",
    )
    compile_parser.add_argument(
        "--app-id",
        help="app id for the connector, from the connector's info response",
        required=True,
    )
    compile_parser.add_argument(
        "--connector-root-module-dir",
        help="root module directory for the connector that contains a main.py. The parent directory should have a pyproject.toml",
        required=True,
    )
    compile_parser.add_argument(
        "--output-directory",
        help="Directory for compilation",
    )

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return
    elif args.command == "scaffold":
        scaffold(args)
        return
    elif args.command == "spec":
        spec_file_path = Path(os.path.dirname(__file__)) / "spec" / "openapi.yaml"
        print(spec_file_path.read_text())
        return
    elif args.command == "compile-on-prem":
        compile_executable_for_onprem(
            connector_root_module_dir=Path(args.connector_root_module_dir),
            app_id=args.app_id,
            exclude_modules=DEFAULT_EXCLUDE_MODULES,
            sdk_root=Path(os.path.dirname(__file__)).parent,
            compile_directory=Path(args.output_directory) if args.output_directory else None,
        )
        return


if __name__ == "__main__":
    main()



import typing as t
from abc import abstractmethod

from connector.httpx_rewrite import AsyncClient
from connector.oai.capability import Request


class BaseIntegrationClient:
    @classmethod
    @abstractmethod
    def prepare_client_args(cls, args: Request) -> dict[str, t.Any]:
        pass

    @classmethod
    def build_client(cls, args) -> AsyncClient:
        return AsyncClient(**cls.prepare_client_args(args))

    def __init__(self, args: Request) -> None:
        self._http_client = self.build_client(args)

    async def __aenter__(self):
        await self._http_client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._http_client.__aexit__()
        if exc_type is not None:
            raise exc_val



"""Tests for ``connector.oai.capability`` module.

Todo
----
* generate_capability_schema
"""

import typing as t

import pytest
import pytest_cases
from connector.oai.capability import (
    CapabilityCallableProto,
    Request,
    Response,
    StandardCapabilityName,
    get_capability_annotations,
    validate_capability,
)

Case: t.TypeAlias = tuple[
    CapabilityCallableProto[t.Any],
    tuple[type[Request], type[Response]],
]


@pytest_cases.parametrize_with_cases(
    ["capability", "expected_annotations"],
    cases=[
        "tests.oai.test_get_capability_annotations_cases",
    ],
    has_tag="correct",
)
async def test_get_capability_annotations(
    capability: CapabilityCallableProto[t.Any],
    expected_annotations: tuple[Request, Response],
) -> None:
    """Test if annotations are correctly get from capability."""
    actual_annotations = get_capability_annotations(capability)
    assert actual_annotations == expected_annotations


@pytest_cases.parametrize_with_cases(
    ["capability"],
    cases=[
        "tests.oai.test_get_capability_annotations_cases",
    ],
    has_tag="missing_annotation",
)
async def test_get_capability_annotations_type_error(
    capability: CapabilityCallableProto[t.Any],
) -> None:
    """Test if incorrectly typed capability raises error."""
    with pytest.raises(TypeError):
        get_capability_annotations(capability)


@pytest_cases.parametrize_with_cases(
    ["capability_name", "capability"],
    cases=[
        "tests.oai.test_validate_capability_cases",
    ],
    has_tag="valid",
)
async def test_validate_capability(
    capability_name: StandardCapabilityName,
    capability: CapabilityCallableProto[t.Any],
) -> None:
    """Test if valid capability is marked as valid."""
    validate_capability(capability_name, capability)


@pytest_cases.parametrize_with_cases(
    ["capability_name", "capability"],
    cases=[
        "tests.oai.test_validate_capability_cases",
    ],
    has_tag="invalid",
)
async def test_validate_capability_invalid(
    capability_name: StandardCapabilityName,
    capability: CapabilityCallableProto[t.Any],
) -> None:
    """Test if valid capability is marked as valid."""
    with pytest.raises(TypeError):
        validate_capability(capability_name, capability)


@pytest_cases.parametrize_with_cases(
    ["capability"],
    cases=[
        "tests.oai.test_get_capability_annotations_cases",
    ],
    has_tag="missing_annotation",
)
async def test_validate_capability_missing_annotation(
    capability: CapabilityCallableProto[t.Any],
) -> None:
    """Test if valid capability is marked as valid.

    We just pass any capability_name just to make function happy,
    however, ``validate_capability`` should raise before it touches the
    name.
    """
    capability_name = StandardCapabilityName.VALIDATE_CREDENTIALS
    with pytest.raises(TypeError):
        validate_capability(capability_name, capability)



"""Test cases for ``Integration.dispatch`` function.

Those cases should lead to raised error.
"""

import json

from connector.generated import (
    BasicCredential,
    Error,
    ErrorCode,
    ErrorResponse,
    ListAccountsRequest,
    ListAccountsResponse,
    StandardCapabilityName,
)
from connector.oai.integration import DescriptionData, Integration
from pydantic import BaseModel

Case = tuple[
    Integration,
    StandardCapabilityName,
    str,
    BaseModel,
]


class SettingsFixture(BaseModel):
    host: str


def new_integration(
    *,
    settings_model: type[BaseModel] | None,
) -> Integration:
    return Integration(
        app_id="test",
        version="0.1.0",
        auth=BasicCredential,
        exception_handlers=[],
        handle_errors=True,
        settings_model=settings_model,
        description_data=DescriptionData(
            user_friendly_name="testing thing",
            categories=[],
        ),
    )


def case_valid_settings_are_fine() -> Case:
    """If somebody passes good settings, we should dispatch as normal"""
    integration = new_integration(settings_model=SettingsFixture)

    @integration.register_capability(StandardCapabilityName.LIST_ACCOUNTS)
    def list_accounts(
        args: ListAccountsRequest,
    ) -> ListAccountsResponse:
        return ListAccountsResponse(response=[])

    capability_name = StandardCapabilityName.LIST_ACCOUNTS
    request = json.dumps(
        {
            "auth": {"basic": {"username": "foo", "password": "bar"}},
            "request": {},
            "settings": {"host": "https://not-checked"},
        }
    )
    expected_response = ListAccountsResponse(response=[])
    return integration, capability_name, request, expected_response


def case_no_settings_are_fine_for_no_model() -> Case:
    """If somebody passes _no_ settings and there's no model, we should dispatch as normal

    ...until Settings are required"""
    integration = new_integration(settings_model=None)

    @integration.register_capability(StandardCapabilityName.LIST_ACCOUNTS)
    def list_accounts(
        args: ListAccountsRequest,
    ) -> ListAccountsResponse:
        return ListAccountsResponse(response=[])

    capability_name = StandardCapabilityName.LIST_ACCOUNTS
    request = json.dumps(
        {
            "auth": {"basic": {"username": "foo", "password": "bar"}},
            "request": {},
        }
    )
    expected_response = ListAccountsResponse(response=[])
    return integration, capability_name, request, expected_response


def case_invalid_settings_are_error() -> Case:
    """If the caller passes an invalid settings object, dispatching should return an error."""
    integration = new_integration(settings_model=SettingsFixture)

    @integration.register_capability(StandardCapabilityName.LIST_ACCOUNTS)
    def list_accounts(
        args: ListAccountsRequest,
    ) -> ListAccountsResponse:
        return ListAccountsResponse(response=[])

    capability_name = StandardCapabilityName.LIST_ACCOUNTS
    request = json.dumps(
        {
            "auth": {"basic": {"username": "foo", "password": "bar"}},
            "request": {},
            "settings": {"something": "incorrect"},
        }
    )
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            error_code=ErrorCode.BAD_REQUEST,
            message="Invalid settings passed on request",
            raised_by=None,
            raised_in=None,
            status_code=None,
            app_id="test",
        ),
    )
    return integration, capability_name, request, expected_response



"""Test related utilities"""


def http_error_message(url: str, status_code: int) -> str:
    """Generate default HTTP error messages from URI and status code"""
    client_error = ""
    match status_code:
        case 400:
            client_error = "Bad Request"
        case 401:
            client_error = "Unauthorized"
        case 404:
            client_error = "Not Found"
        case 429:
            client_error = "Too Many Requests"
        case _:
            pass

    return f"Client error '{status_code} {client_error}' for url '{url}'\nFor more information check: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/{status_code}"



"""Cases for testing ``activate_account`` capability."""

import typing as t

import httpx
from connector.generated import (
    AccountStatus,
    ActivateAccount,
    ActivateAccountRequest,
    ActivateAccountResponse,
    ActivatedAccount,
    ErrorResponse,
)
from connector.tests.type_definitions import MockedResponse, ResponseBodyMap

from tests.common_mock_data import SETTINGS, VALID_AUTH

from connector.generated.models.standard_capability_name import StandardCapabilityName

Case: t.TypeAlias = tuple[
    StandardCapabilityName,
    ActivateAccountRequest,
    ResponseBodyMap,
    ActivateAccountResponse | ErrorResponse,
]


def case_activate_account_200() -> Case:
    """Activate Account - Successful case."""
    args = ActivateAccountRequest(
        request=ActivateAccount(
            account_id="1",
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )

    response_body_map = {{
        "PUT": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ActivateAccountResponse(
        response=ActivatedAccount(activated=True, status=AccountStatus.ACTIVE),
    )
    return StandardCapabilityName.ACTIVATE_ACCOUNT, args, response_body_map, expected_response



import datetime
import logging
from pathlib import Path

from connector.config import config

logger = logging.getLogger(__name__)


def set_logger_config(app_id: str):
    log_directory = config.log_directory
    log_level = config.log_level

    timestamp = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d")

    if log_directory:
        log_directory.mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            filename=str(log_directory / f"{app_id}_{timestamp}.log"),
            filemode="a",
            format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
            datefmt="%H:%M:%S",
            level=log_level.value,
            force=True,
        )
    else:
        log_directory = Path.cwd() / "logs"
        log_directory.mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            filename=str(log_directory / f"{app_id}_{timestamp}.log"),
            format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
            datefmt="%H:%M:%S",
            level=log_level.value,
            force=True,
        )



from connector.generated import (
    FindEntitlementAssociationsRequest,
    FindEntitlementAssociationsResponse,
    FoundAccountData,
    GetLastActivityRequest,
    GetLastActivityResponse,
    ListAccountsRequest,
    ListAccountsResponse,
    ListEntitlementsRequest,
    ListEntitlementsResponse,
    ListResourcesRequest,
    ListResourcesResponse,
    Page,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from connector.oai.capability import get_page, get_settings


from {name}.pagination import DEFAULT_PAGE_SIZE, NextPageToken, Pagination
from {name}.settings import {pascal}Settings
from {name}.client import {pascal}Client

async def validate_credentials(
    args: ValidateCredentialsRequest,
) -> ValidateCredentialsResponse:
    async with {pascal}Client(args) as client:
        #_ = await client.get_users()

        return ValidateCredentialsResponse(
            response=ValidatedCredentials(
                unique_tenant_id="REPLACE_WITH_UNIQUE_TENANT_ID",
                valid=True,
            ),
        )


async def list_accounts(args: ListAccountsRequest) -> ListAccountsResponse:
    endpoint = "/users"
    try:
        current_pagination = NextPageToken(get_page(args).token).paginations()[0]
    except IndexError:
        current_pagination = Pagination.default(endpoint)

    page_size = get_page(args).size or DEFAULT_PAGE_SIZE
    async with {pascal}Client(args) as client:
        response = await client.get_users(
            limit=page_size, offset=current_pagination.offset,
        )
        accounts: list[FoundAccountData] = response.to_accounts()

        next_pagination = []
        if True:
            next_pagination.append(
                Pagination(
                    endpoint=endpoint,
                    offset=current_pagination.offset + len(accounts),
                )
            )

        next_page_token = NextPageToken.from_paginations(next_pagination).token

    return ListAccountsResponse(
        response=accounts,
        page=Page(
            token=next_page_token,
            size=page_size,
        )
        if next_page_token
        else None,
    )


async def list_resources(args: ListResourcesRequest) -> ListResourcesResponse:
    raise NotImplementedError


async def list_entitlements(
    args: ListEntitlementsRequest,
) -> ListEntitlementsResponse:
    raise NotImplementedError


async def find_entitlement_associations(
    args: FindEntitlementAssociationsRequest,
) -> FindEntitlementAssociationsResponse:
    raise NotImplementedError


async def get_last_activity(args: GetLastActivityRequest) -> GetLastActivityResponse:
    raise NotImplementedError



"""Cases for testing ``delete_account`` capability."""

import typing as t

import httpx
from connector.generated import (
    AccountStatus,
    DeleteAccount,
    DeleteAccountRequest,
    DeleteAccountResponse,
    DeletedAccount,
    Error,
    ErrorCode,
    ErrorResponse,
)
from connector.utils.test import http_error_message

from tests.common_mock_data import INVALID_AUTH, SETTINGS, VALID_AUTH
from connector.tests.type_definitions import MockedResponse, ResponseBodyMap

from connector.generated.models.standard_capability_name import StandardCapabilityName

Case: t.TypeAlias = tuple[
    StandardCapabilityName,
    DeleteAccountRequest,
    ResponseBodyMap,
    DeleteAccountResponse | ErrorResponse,
]


def case_delete_account_204() -> Case:
    """Successful deletion request."""
    args = DeleteAccountRequest(
        request=DeleteAccount(
            account_id="1",
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "DELETE": {{
            f"/users/{{args.request.account_id}}": MockedResponse(
                status_code=httpx.codes.NO_CONTENT,
                response_body=None,
            ),
        }},
    }}
    expected_response = DeleteAccountResponse(
        response=DeletedAccount(deleted=True, status=AccountStatus.DELETED),
    )
    return StandardCapabilityName.DELETE_ACCOUNT, args, response_body_map, expected_response


def case_delete_account_404() -> Case:
    """User not found request should fail."""
    args = DeleteAccountRequest(
        request=DeleteAccount(
            account_id="non_existent",
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "DELETE": {{
            f"/users/{{args.request.account_id}}": MockedResponse(
                status_code=httpx.codes.NOT_FOUND,
                response_body={{
                    "error": {{
                        "message": "Not found",
                        "code": 2100,
                    }},
                }},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                "",
                httpx.codes.NOT_FOUND,
            ),
            status_code=httpx.codes.NOT_FOUND,
            error_code=ErrorCode.NOT_FOUND,
            app_id="{name}",
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:delete_account",
        ),
    )
    return StandardCapabilityName.DELETE_ACCOUNT, args, response_body_map, expected_response
